import logging

from django.core.exceptions import ValidationError
from django.core.validators import URLValidator
from django.http import HttpResponseBadRequest  # HTTP 400
from django.http import HttpResponse, JsonResponse
from django.shortcuts import render
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import View

from mentions.exceptions import TargetDoesNotExist
from mentions.forms.manual_submit_webmention import ManualSubmitWebmentionForm
from mentions.tasks import process_incoming_webmention
from mentions.util import get_mentions_for_url_path, serialize_mentions, split_url

log = logging.getLogger(__name__)


def _get_client_ip(request):
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
    if x_forwarded_for:
        ip = x_forwarded_for.split(",")[-1].strip()
    else:
        ip = request.META.get("REMOTE_ADDR")
    return ip


def _validate(validator, urls):
    for url in urls:
        try:
            validator(url)
        except ValidationError as e:
            log.warning(f"URL '{url}' did not pass validation: {e}")
            return False
    return True


# /webmention/
class WebmentionView(View):
    """Handle incoming webmentions."""

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def get(self, request):
        form = ManualSubmitWebmentionForm()
        return render(request, "webmention-submit-manual.html", {"form": form})

    def post(self, request):
        log.info("Receiving webmention...")
        http_post = request.POST

        try:
            client_ip = _get_client_ip(request)
            source = http_post["source"]
            target = http_post["target"]

        except Exception as e:
            log.warning(f'Unable to read webmention params "{http_post}": {e}')
            return HttpResponse(status=400)

        validator = URLValidator(schemes=["http", "https"])

        if _validate(validator, [source, target]):
            log.info(f"Validation passed for source: '{source}', target: '{target}'")
        else:
            return HttpResponseBadRequest()

        process_incoming_webmention.delay(http_post, client_ip)
        return HttpResponse("Thank you, your webmention has been accepted.", status=202)


# /webmention/get
class GetWebmentionsView(View):
    """Return any webmentions associated with a given url."""

    def get(self, request):
        for_url = request.GET.get("url")

        if not for_url:
            return HttpResponseBadRequest("Missing args")

        scheme, domain, path = split_url(request.build_absolute_uri())
        full_target_url = f"{scheme}://{domain}{for_url}"

        try:
            wm = get_mentions_for_url_path(for_url, full_target_url=full_target_url)

            return JsonResponse(
                {
                    "target_url": full_target_url,
                    "mentions": serialize_mentions(wm),
                }
            )

        except TargetDoesNotExist as e:
            log.warning(e)
            return JsonResponse(
                {
                    "target_url": full_target_url,
                    "message": "Target not found",
                    "mentions": [],
                },
                status=404,
            )



from django.test import TestCase


class WebmentionTestCase(TestCase):
    def tearDown(self) -> None:
        super().tearDown()
        from mentions.models import (
            HCard,
            OutgoingWebmentionStatus,
            SimpleMention,
            Webmention,
        )

        for Model in [
            Webmention,
            OutgoingWebmentionStatus,
            HCard,
            SimpleMention,
        ]:
            Model.objects.all().delete()


class MockResponse:
    """Mock of requests.Response."""

    def __init__(
        self,
        url: str,
        headers: dict = None,
        text: str = None,
        status_code: int = None,
    ):
        self.url = url
        self.text = text
        self.status_code = status_code
        self.headers = headers



import requests
from celery import shared_task
from celery.utils.log import get_task_logger
from django.conf import settings
from django.db import models
from django.http import QueryDict

from mentions.exceptions import (
    BadConfig,
    SourceNotAccessible,
    TargetDoesNotExist,
    TargetWrongDomain,
)
from mentions.models import HCard, Webmention
from mentions.util import get_model_for_url_path, html_parser, split_url

log = get_task_logger(__name__)


class Notes:
    notes = []

    def info(self, note) -> "Notes":
        log.info(note)
        self.notes.append(note)
        return self

    def warn(self, note) -> "Notes":
        log.warning(note)
        self.notes.append(note)
        return self

    def join_to_string(self):
        return "\n".join(self.notes)


def _update_wm(
    mention,
    target_object=None,
    notes: Notes = None,
    hcard: HCard = None,
    validated: bool = None,
    save: bool = False,
):
    """Update a webmention with the given kwargs"""
    if target_object is not None:
        mention.target_object = target_object
    if notes is not None:
        mention.notes = notes.join_to_string()[:1000]
    if hcard is not None:
        mention.hcard = hcard
    if validated is not None:
        mention.validated = validated

    if save:
        log.info(f"Webmention saved: {mention}")
        mention.save()

    return mention


@shared_task
def process_incoming_webmention(http_post: QueryDict, client_ip: str) -> None:
    log.info(f"Processing webmention '{http_post}'")

    # Source and target have already been verified
    # as valid addresses before this method is called
    source = http_post["source"]
    target = http_post["target"]

    wm = Webmention(source_url=source, target_url=target, sent_by=client_ip)

    # If anything fails, write it to notes and attach to webmention object
    # so it can be checked later
    notes = Notes()

    # Check that the target page is accessible on our server and fetch
    # the corresponding object.
    try:
        obj = _get_target_object(target)
        notes.info("Found webmention target object")
        _update_wm(wm, target_object=obj)

    except (TargetWrongDomain, TargetDoesNotExist):
        notes.warn(f"Unable to find matching page on our server for url '{target}'")

    except BadConfig:
        notes.warn(f"Unable to find a model associated with url '{target}'")

    # Verify that the source page exists and really contains a link
    # to the target
    try:
        response_text = _get_incoming_source(source)

    except SourceNotAccessible:
        _update_wm(
            wm, notes=notes.warn(f"Source not accessible: '{source}'"), save=True
        )
        return

    soup = html_parser(response_text)
    if not soup.find("a", href=target):
        _update_wm(
            wm,
            notes=notes.info("Source does not contain a link to our content"),
            save=True,
        )
        return

    hcard = HCard.from_soup(soup, save=True)

    _update_wm(wm, validated=True, notes=notes, hcard=hcard, save=True)


def _get_target_object(target_url: str) -> models.Model:
    r"""
    Confirm that the page exists on our server and return object.
    Throws

    Any mentionable Views must include the following kwargs:
        - 'model_name' - dotted python path to the model definition
                         This needs to be explicitely added to kwargs
        - 'slug'       - normally parsed from url pattern, used to
                         resolve the particular model instance

    e.g.
    re_path(
        r'^articles/(?P<slug>[\w\-]+)/?$',
        ArticleView.as_view(),
        kwargs={
            'model_name': 'core.models.Article',
        },
        name='app_view'),

    Raises:
        TargetWrongDomain: If the target_url points to a domain not listed in settings.ALLOWED_HOSTS
        BadConfig: Raised from get_model_for_url
    """
    scheme, domain, path = split_url(target_url)

    if domain not in settings.ALLOWED_HOSTS:
        raise TargetWrongDomain(f"Wrong domain: {domain} (from url={target_url})")

    try:
        return get_model_for_url_path(path)
    except BadConfig as e:
        raise e


def _get_incoming_source(source_url: str, client=requests) -> str:
    """
    Fetch the source, confirm content is suitable and return response.
    Verify that the source URL returns an HTML page with a successful
    status code.

    Args:
        source_url: The URL that mentions our content.
        client: A client for making HTTP requests. Should only be explicitly given when testing -
                its API is assumed to be equivalent to `python-requests`.

    Raises:
        SourceNotAccessible: If the `source_url` cannot be resolved, returns an error code, or
                             is an unexpected content type.
    """

    try:
        response = client.get(source_url)
    except Exception as e:
        raise SourceNotAccessible(f"Requests error: {e}")

    if response.status_code >= 300:
        raise SourceNotAccessible(
            f"Source '{source_url}' returned error code [{response.status_code}]"
        )

    content_type = response.headers["content-type"]
    if "text/html" not in content_type:
        raise SourceNotAccessible(
            f"Source '{source_url}' returned unexpected content type: {content_type}"
        )

    return response.text



import logging

from django.urls import path

from mentions import util
from mentions.exceptions import BadConfig, TargetDoesNotExist
from tests import WebmentionTestCase
from tests.models import MentionableTestModel
from tests.util import constants, testfunc
from tests.views import AllEndpointsMentionableTestView

log = logging.getLogger(__name__)

# Paths as configured in local_urlpatterns
bad_modelname_key = "with_bad_kwarg_key"
bad_modelname_value = "with_bad_kwarg_value"
bad_path = "some_unresolvable_slug"

"""Url patterns that are only used for tests in this file.
Should be added to test_urls.urlpatterns in setUp and removed again in tearDown."""
local_urlpatterns = [
    path(
        fr"{bad_modelname_key}/<slug:slug>",
        AllEndpointsMentionableTestView.as_view(),
        kwargs={
            "model_name_with_mistyped_or_missing_key": constants.model_name,
        },
    ),
    path(
        fr"{bad_modelname_value}/<slug:slug>",
        AllEndpointsMentionableTestView.as_view(),
        kwargs={
            "model_name": "tests.UnresolvableModel",
        },
    ),
]


class _BaseTestCase(WebmentionTestCase):
    def setUp(self):
        self.pk, self.slug = testfunc.get_id_and_slug()

        _create_mentionable_objects()
        MentionableTestModel.objects.create(pk=self.pk, slug=self.slug)
        _create_mentionable_objects()


class GetModelForUrlPathTests(_BaseTestCase):
    """Tests for util.get_model_for_url_path with correct `urlpatterns` configuration."""

    def test_get_model_for_url__with_correct_slug(self):
        """Reverse URL lookup finds the correct object."""
        retrieved_object = util.get_model_for_url_path(testfunc.get_urlpath(self.slug))

        self.assertEqual(retrieved_object.pk, self.pk)

    def test_get_model_for_url__with_unknown_url(self):
        """URL with an unrecognised model_name path raises TargetDoesNotExist exception."""
        with self.assertRaises(TargetDoesNotExist):
            util.get_model_for_url_path("/some/nonexistent/urlpath")

    def test_get_model_for_url__with_unknown_slug(self):
        """Unknown slug raises TargetDoesNotExist exception."""
        with self.assertRaises(TargetDoesNotExist):
            util.get_model_for_url_path(testfunc.get_urlpath("unknown-slug"))


class GetModelForUrlPathWithBadConfigTests(_BaseTestCase):
    """Tests for util.get_model_for_url_path when there are errors in urlpatterns path configuration."""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        from tests.config.test_urls import urlpatterns

        urlpatterns += local_urlpatterns

    def test_get_model_for_url__with_bad_model_name_config(self):
        """urlpatterns with no entry for model_name in path kwargs raises BadConfig exception."""
        with self.assertRaises(BadConfig):
            util.get_model_for_url_path(_urlpath(bad_modelname_key, self.slug))

    def test_get_model_for_url__raises_badconfig_when_model_name_unresolvable(self):
        """Unresolvable model_name raises BadConfig exception."""
        with self.assertRaises(BadConfig):
            util.get_model_for_url_path(_urlpath(bad_modelname_value, self.slug))

    @classmethod
    def tearDownClass(cls) -> None:
        from tests.config.test_urls import urlpatterns

        for x in local_urlpatterns:
            urlpatterns.remove(x)

        super().tearDownClass()


def _urlpath(path: str, slug: str) -> str:
    return f"/{path}/{slug}"


def _create_mentionable_objects(n: int = 3):
    """Create some arbitrary mentionable objects for noise."""

    for x in range(n):
        pk, slug = testfunc.get_id_and_slug()
        MentionableTestModel.objects.create(pk=pk, slug=slug)



import logging
from importlib import import_module
from typing import Dict, Iterable, List, Optional, Tuple
from urllib.parse import urlsplit

from bs4 import BeautifulSoup
from django.apps import apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q, QuerySet
from django.urls import Resolver404, ResolverMatch

from mentions.exceptions import BadConfig, TargetDoesNotExist
from mentions.models import HCard, Webmention

log = logging.getLogger(__name__)


def _find_urlpattern(target_path: str) -> ResolverMatch:
    """
    Find a match in urlpatterns or raise TargetDoesNotExist
    """
    target_path = target_path.lstrip("/")  # Remove any leading slashes
    urlconf = import_module(settings.ROOT_URLCONF)
    urlpatterns = urlconf.urlpatterns

    try:
        from django.contrib.flatpages.views import flatpage
    except ImportError:
        flatpage = None

    for x in urlpatterns:
        # x may be an instance of either:
        # - django.urls.resolvers.URLResolver
        # - django.urls.resolvers.URLPattern
        try:
            match = x.resolve(target_path)
            if match:
                if flatpage and match.func != flatpage:
                    break
                else:
                    break
        except Resolver404:
            # May be raised by URLResolver
            pass
    else:
        # No match found
        raise TargetDoesNotExist(
            f"Cannot find a matching urlpattern entry for path={target_path}"
        )
    log.debug(f"Found matching urlpattern: {match}")
    return match


def split_url(target_url: str) -> Tuple[str, str, str]:
    scheme, full_domain, path, _, _ = urlsplit(target_url)
    domain = full_domain.split(":")[0]  # Remove port number if present
    return scheme, domain, path


def get_model_for_url_path(target_path: str, match: ResolverMatch = None):
    """
    Find a match in urlpatterns and return the corresponding model instance.
    """
    if match is None:
        match = _find_urlpattern(target_path)

    # Dotted path to model class declaration
    model_name = match.kwargs.get("model_name")

    # Required to retrieve the target model instance
    slug = match.kwargs.get("slug")

    if not model_name:
        raise BadConfig(
            f"urlpattern must include a kwarg entry called 'model_name': {match}"
        )
    if not slug:
        raise BadConfig(f"urlpattern must include a kwarg entry called 'slug': {match}")

    try:
        model = apps.get_model(model_name)
    except LookupError:
        raise BadConfig(f"Cannot find model `{model_name}` - check your urlpattern!")

    try:
        return model.objects.get(slug=slug)
    except ObjectDoesNotExist:
        raise TargetDoesNotExist(
            f"Cannot find instance of model='{model}' with slug='{slug}'"
        )


def get_mentions_for_url_path(
    target_path: str,
    full_target_url: str,
) -> "QuerySet[Webmention]":
    """If target_path resolves to a page associated with a MentionableMixin model"""
    match = _find_urlpattern(target_path)

    try:
        obj = get_model_for_url_path(target_path, match)
        return obj.mentions
    except (BadConfig, TargetDoesNotExist):
        pass

    # Add or remove a trailing slash to full_target_url so we can look for both.
    full_target_url_invert_slash = (
        full_target_url[:-1] if full_target_url.endswith("/") else f"{full_target_url}/"
    )

    mentions = Webmention.objects.filter(
        Q(target_url=full_target_url)
        | Q(target_url=full_target_url_invert_slash)
        | Q(target_url=target_path),
        approved=True,
        validated=True,
    )

    return mentions


def serialize_mentions(mentions: Iterable[Webmention]) -> List[Dict]:
    return [
        {
            "hcard": serialize_hcard(mention.hcard),
            "quote": mention.quote,
            "source_url": mention.source_url,
            "published": mention.published,
        }
        for mention in mentions
    ]


def serialize_hcard(hcard: Optional[HCard]) -> Optional[Dict]:
    if hcard is None:
        return None

    return {
        "name": hcard.name,
        "avatar": hcard.avatar,
        "homepage": hcard.homepage,
    }


def html_parser(content) -> BeautifulSoup:
    return BeautifulSoup(content, features="html5lib")



"""
Utility functions used in multiple test files.
"""
import random
import uuid
from typing import Tuple

from django.conf import settings
from django.urls import reverse
from django.utils.text import slugify

from mentions.views import view_names

from . import viewname


def get_id_and_slug() -> Tuple[int, str]:
    """Create a random id and slug for a MentionableTestModel."""
    _id = random.randint(1, 1_000_000)
    slug = slugify(_id)
    return _id, slug


def get_url(slug):
    """Return the absolute URL for the target slug on our domain."""
    return f"https://{settings.DOMAIN_NAME}{get_urlpath(slug)}"


def get_urlpath(slug: str) -> str:
    """Return the relative URL for the target slug."""
    return reverse(viewname.with_all_endpoints, args=[slug])


def endpoint_submit_webmention() -> str:  #
    """Return relative URL for our root webmention endpoint."""
    return reverse(view_names.webmention_api_incoming)


def endpoint_get_webmentions() -> str:
    """Return relative URL for our webmention /get endpoint."""
    return reverse(view_names.webmention_api_get_for_object)


def endpoint_submit_webmention_absolute() -> str:
    """Return absolute URL for our root webmention endpoint on our domain."""
    return f"https://{settings.DOMAIN_NAME}{endpoint_submit_webmention()}"


def random_domain() -> str:
    """Return a randomised domain name."""
    return f"example-url-{uuid.uuid4().hex[:5]}.org"



import logging

from django.contrib.contenttypes.models import ContentType
from django.db import models

from mentions.exceptions import ImplementationRequired
from mentions.models.webmention import Webmention
from mentions.tasks import process_outgoing_webmentions
from mentions.util import serialize_mentions

log = logging.getLogger(__name__)


class MentionableMixin(models.Model):
    class Meta:
        abstract = True

    allow_incoming_webmentions = models.BooleanField(default=True)
    allow_outgoing_webmentions = models.BooleanField(default=False)

    slug = models.SlugField(unique=True)

    @property
    def mentions(self):
        ctype = ContentType.objects.get_for_model(self.__class__)
        webmentions = Webmention.objects.filter(
            content_type=ctype,
            object_id=self.id,
            approved=True,
            validated=True,
        )
        # manual_mentions = TODO
        return webmentions

    def mentions_json(self):
        return serialize_mentions(self.mentions)

    def get_absolute_url(self):
        raise ImplementationRequired(
            f"{self.__class__.__name__} does not implement get_absolute_url()"
        )

    def all_text(self) -> str:
        """
        Return all the text that should be searched when looking for
        outgoing Webmentions. Any URLs found in this text will be
        checked for webmention support.

        Typically this will just be the main text of your model but
        you may also want to include content from any other text fields
        such as a summary or abstract.

        Example:
            def all_text(self) -> str:
                return f'{self.introduction} {self.main_content}'
        """
        raise ImplementationRequired(
            f"{self.__class__.__name__} does not implement all_text()"
        )

    def save(self, *args, **kwargs):
        if self.allow_outgoing_webmentions:
            log.info("Outgoing webmention processing task added to queue...")
            process_outgoing_webmentions.delay(self.get_absolute_url(), self.all_text())

        super().save(*args, **kwargs)



import json
import logging
from functools import reduce
from typing import Optional

import mf2py
from bs4 import BeautifulSoup
from django.db import models

from mentions.exceptions import NotEnoughData
from mentions.models import MentionsBaseModel

log = logging.getLogger(__name__)


class HCard(MentionsBaseModel):
    """An h-card represents information about a person or organisation.

    Spec: https://microformats.org/wiki/h-card
    """

    name = models.CharField(
        max_length=50,
        blank=True,
        help_text="Name of the person/organisation",
    )
    avatar = models.URLField(blank=True, help_text="Link to their profile image")
    homepage = models.URLField(
        blank=True, null=True, help_text="Link to their homepage"
    )
    json = models.TextField(
        blank=True, help_text="Raw json representation of this hcard"
    )

    @classmethod
    def from_soup(cls, soup: BeautifulSoup, save=False):
        """Create or update HCard(s) using data from a BeautifulSoup document.

        See https://github.com/microformats/mf2py"""
        parser = mf2py.Parser(doc=soup)
        parsed_data = parser.to_dict()
        for item in parsed_data.get("items", []):
            try:
                return _parse_hcard(item, save)
            except NotEnoughData as e:
                log.debug(e)
                continue

    def save(self, *args, **kwargs):
        """Workaround so that empty homepages can be accepted without violating
        'unique' requirement. This allows admin to check missing content manually."""
        if not self.homepage:
            self.homepage = None

        super().save(*args, **kwargs)

    def __str__(self):
        return f'name="{self.name}", avatar="{self.avatar}", homepage="{self.homepage}"'


def _parse_hcard(item: dict, save: bool) -> Optional[HCard]:
    """Spec: https://github.com/microformats/mf2py"""
    if "h-card" in item.get("type"):
        log.debug("Reading h-card...")
        attrs = item.get("properties")
        if attrs:
            _name = attrs.get("name", [""])[0]
            _avatar = attrs.get("photo", [""])[0]
            _homepage = attrs.get("url", [None])[0]
            _json = json.dumps(item, sort_keys=True)

            require_one_of = [_name, _avatar, _homepage]
            has_required_fields = (
                reduce(lambda acc, value: acc + 1 if value else acc, require_one_of, 0)
                >= 1
            )

            if not has_required_fields:
                raise NotEnoughData(
                    "An HCard requires a value for at least one of 'name', 'avatar', 'homepage'"
                )

            card, _ = HCard.objects.get_or_create(homepage=_homepage)

            card.name = _name
            card.avatar = _avatar
            card.json = _json

            if save:
                card.save()

            return card
        else:
            log.debug('_parse_hcard could not read "properties"')



from django import template
from django.urls import reverse
from django.utils.html import format_html

from mentions.views import view_names

register = template.Library()


@register.simple_tag
def webmention_endpoint():
    """Add your webmentions endpoint to your template <head>.

    e.g.
    {% load webmention_endpoint %}
    ...
    <head>
        {% webmention_endpoint %}
    </head>
    ...
    """
    endpoint = reverse(view_names.webmention_api_incoming)
    return format_html("""<link rel="webmention" href="{}" />""", endpoint)



from django.conf import settings

from mentions.tasks import outgoing_webmentions
from tests import MockResponse, WebmentionTestCase
from tests.util import snippets, testfunc

OUTGOING_WEBMENTION_HTML_DUPLICATE_LINKS = f"""<html>
<head><link rel="webmention" href="/webmention/" /></head>
<body>blah blah 
<a href="https://beatonma.org/">This is a mentionable link</a> 
blah blah duplicate links
<a href="https://snommoc.org/">This is some other link</a> 
<a href="https://beatonma.org/">This is a duplicate link</a> 
</body></html>
"""


class EndpointDiscoveryTests(WebmentionTestCase):
    """Outgoing webmentions: Endpoint discovery & resolution"""

    absolute_endpoint = testfunc.endpoint_submit_webmention_absolute()
    relative_endpoint = testfunc.endpoint_submit_webmention()

    def setUp(self):
        target_pk, self.target_slug = testfunc.get_id_and_slug()

    def _get_absolute_target_url(self):
        return testfunc.get_url(self.target_slug)

    def test_get_absolute_endpoint_from_response(self):
        """Any exposed endpoints (in HTTP header, HTML <head> or <body>) are found and returned as an absolute url."""
        mock_response = MockResponse(
            url=self._get_absolute_target_url(),
            headers={"Link": snippets.http_link_endpoint()},
        )
        absolute_endpoint_from_http_headers = (
            outgoing_webmentions._get_absolute_endpoint_from_response(mock_response)
        )
        self.assertEqual(self.absolute_endpoint, absolute_endpoint_from_http_headers)

        mock_response.headers = {}
        mock_response.text = snippets.html_head_endpoint()
        absolute_endpoint_from_html_head = (
            outgoing_webmentions._get_absolute_endpoint_from_response(mock_response)
        )
        self.assertEqual(self.absolute_endpoint, absolute_endpoint_from_html_head)

        mock_response.headers = {}
        mock_response.text = snippets.html_body_endpoint()
        absolute_endpoint_from_html_body = (
            outgoing_webmentions._get_absolute_endpoint_from_response(mock_response)
        )
        self.assertEqual(self.absolute_endpoint, absolute_endpoint_from_html_body)

    def test_get_endpoint_in_http_headers(self):
        """Endpoints exposed in HTTP header are found correctly."""

        mock_response = MockResponse(
            url=self._get_absolute_target_url(),
            headers={"Link": snippets.http_link_endpoint()},
        )
        endpoint_from_http_headers = outgoing_webmentions._get_endpoint_in_http_headers(
            mock_response
        )
        self.assertEqual(self.relative_endpoint, endpoint_from_http_headers)

    def test_get_endpoint_in_html_head(self):
        """Endpoints exposed in HTML <head> are found correctly."""

        mock_response = MockResponse(
            url=self._get_absolute_target_url(),
            text=snippets.html_head_endpoint(),
        )
        endpoint_from_html_head = outgoing_webmentions._get_endpoint_in_html_response(
            mock_response
        )
        self.assertEqual(self.relative_endpoint, endpoint_from_html_head)

    def test_get_endpoint_in_html_body(self):
        """Endpoints exposed in HTML <body> are found correctly."""

        mock_response = MockResponse(
            url=self._get_absolute_target_url(),
            text=snippets.html_body_endpoint(),
        )
        endpoint_from_html_body = outgoing_webmentions._get_endpoint_in_html_response(
            mock_response
        )
        self.assertEqual(self.relative_endpoint, endpoint_from_html_body)

    def test_relative_to_absolute_url(self):
        """Relative URLs are correctly converted to absolute URLs."""

        domain = settings.DOMAIN_NAME
        page_url = f"https://{domain}/some-url-path"
        response = MockResponse(url=page_url)

        absolute_url_from_root = outgoing_webmentions._relative_to_absolute_url(
            response, "/webmention/"
        )
        self.assertEqual(f"https://{domain}/webmention/", absolute_url_from_root)

        absolute_url_from_relative = outgoing_webmentions._relative_to_absolute_url(
            response, "webmention/"
        )
        self.assertEqual(f"https://{domain}/webmention/", absolute_url_from_relative)

        already_absolute_url = outgoing_webmentions._relative_to_absolute_url(
            response, f"https://{domain}/already_absolute_path"
        )
        self.assertEqual(
            f"https://{domain}/already_absolute_path", already_absolute_url
        )

    def test_find_links_in_text(self):
        """Outgoing links are found correctly."""

        urls = {
            f"https://{testfunc.random_domain()}",
            f"http://{testfunc.random_domain()}/some-path",
            f"{testfunc.random_domain()}/some-path/something_else_04/",
            f"https://subdomain.{testfunc.random_domain()}/blah-blah/",
        }

        outgoing_content = "".join(
            [
                f'This is some content that mentions <a href="{url}">this page</a>'
                for url in urls
            ]
        )

        outgoing_links = outgoing_webmentions._find_links_in_text(outgoing_content)
        self.assertSetEqual(outgoing_links, urls)

    def test_find_links_in_text__should_remove_duplicates(self):
        """_find_links_in_text should remove any duplicates."""
        outgoing_links = outgoing_webmentions._find_links_in_text(
            OUTGOING_WEBMENTION_HTML_DUPLICATE_LINKS
        )

        self.assertSetEqual(
            {"https://beatonma.org/", "https://snommoc.org/"}, outgoing_links
        )



"""
Make sure we can correctly retrieve Webmentions for a given url/object.
"""

import json
import logging
from typing import Dict, List, Optional, Tuple

from django.urls import reverse

from mentions.models import Webmention
from tests import WebmentionTestCase
from tests.models import MentionableTestModel
from tests.util import testfunc, viewname

log = logging.getLogger(__name__)


class _MentionsForObjectTestCase(WebmentionTestCase):
    """Tests for `/get` endpoint"""

    endpoint = testfunc.endpoint_get_webmentions()

    def get_json_response(
        self, url, data=None, **kwargs
    ) -> Tuple[int, Optional[List[Dict]]]:
        response = self.client.get(url, data=data, **kwargs)
        return response.status_code, json.loads(response.content).get("mentions")


class WebmentionGetTests(_MentionsForObjectTestCase):
    """Webmention /get endpoint tests"""

    def setUp(self):
        target_pk, target_slug = testfunc.get_id_and_slug()

        target = MentionableTestModel.objects.create(pk=target_pk, slug=target_slug)
        self.target_urlpath = testfunc.get_urlpath(target_slug)

        _, source_slug = testfunc.get_id_and_slug()
        self.source_url = testfunc.get_url(source_slug)

        Webmention.objects.create(
            source_url=testfunc.get_url(source_slug),
            validated=True,
            approved=True,
            target_object=target,
        )

    def test_get_webmentions_view__with_valid_object(self):
        """Correctly resolve webmentions for a MentionableMixin object from a URL."""
        status, mentions = self.get_json_response(
            self.endpoint, data={"url": self.target_urlpath}
        )

        self.assertEqual(status, 200)
        self.assertEqual(len(mentions), 1)

        first_mention = mentions[0]
        self.assertEqual(first_mention["source_url"], self.source_url)


class WebmentionGetBadRequestTests(_MentionsForObjectTestCase):
    """`get/` endpoint: Poorly-formed requests should give suitable responses"""

    def test_get_webmentions_view__require_http_get(self):
        """get/ endpoint does not accept HTTP POST/PATCH/DELETE."""
        response = self.client.post(self.endpoint)
        self.assertEqual(405, response.status_code)

        response = self.client.patch(self.endpoint)
        self.assertEqual(405, response.status_code)

        response = self.client.delete(self.endpoint)
        self.assertEqual(405, response.status_code)

    def test_get_webmentions_view__require_param_for_url(self):
        """Missing query parameter `url` gives 400 error code."""
        response = self.client.get(self.endpoint)
        self.assertEqual(400, response.status_code)

    def test_get_webmentions_view__target_does_not_exist(self):
        """If target URL does not exist, return 404 error code with empty mentions list."""
        status, mentions = self.get_json_response(
            self.endpoint,
            data={"url": "/does-not-exist"},
        )

        self.assertEqual(status, 404)
        self.assertListEqual([], mentions)


class WebmentionNoModelTests(_MentionsForObjectTestCase):
    """`get/` endpoint: Pages with no associated model should still be mentionable."""

    def setUp(self) -> None:
        self.target_url = reverse(viewname.with_no_mentionable_object)
        Webmention.objects.create(
            source_url="https://beatonma.org/",
            target_url=self.target_url,
            approved=True,
            validated=True,
        )

    def test_get_webmentions_view__no_mentionable_model(self):
        """Webmentions can be retrieved for a URL even if there is not a MentionableMixin model associated with it."""
        status, mentions = self.get_json_response(
            self.endpoint,
            data={"url": self.target_url},
        )

        self.assertEqual(status, 200)
        self.assertEqual(len(mentions), 1)

        first_mention = mentions[0]
        self.assertEqual(first_mention["source_url"], "https://beatonma.org/")



from django.urls import reverse

from mentions.tasks import outgoing_webmentions
from tests import WebmentionTestCase
from tests.util import testfunc


class TemplateTagTests(WebmentionTestCase):
    def test_webmention_endpoint_templatetag(self):
        expected_endpoint = testfunc.endpoint_submit_webmention()
        response = self.client.get(reverse("test-template-tags"))

        self.assertTemplateUsed(response, "templatetags_example.html")
        self.assertContains(response, expected_endpoint)

        self.assertEqual(
            outgoing_webmentions._get_endpoint_in_html(response.content),
            expected_endpoint,
        )



import os
import sys
import uuid


def _python_version_at_least(required_major, required_minor, required_micro=0):
    major, minor, micro, releaselevel, serial = sys.version_info
    return (
        (major >= required_major)
        and (minor >= required_minor)
        and (micro >= required_micro)
    )


if _python_version_at_least(3, 10, 0):
    # NoseTests is not compatible with Python versions >= 3.10
    TEST_RUNNER = "django.test.runner.DiscoverRunner"
else:
    # Pretty print test results with NoseTests
    TEST_RUNNER = "django_nose.NoseTestSuiteRunner"

NOSE_ARGS = [
    "--with-spec",
    "--spec-color",
    "--logging-clear-handlers",
    "--traverse-namespace",  # Required since Python 3.8
    "--exe",
]

# Randomise domain name at test runtime.
DOMAIN_NAME = f"example-url-{uuid.uuid4().hex[::5]}.org"
ALLOWED_HOSTS = [
    DOMAIN_NAME,
]

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

SECRET_KEY = "some-test-key"
INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.flatpages",
    "django.contrib.sites",
    # Apps for django-wm
    "mentions",
    "tests",
]

MIDDLEWARE = [
    # Middleware for django-wm
    "mentions.middleware.WebmentionHeadMiddleware",
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "django-wm-test.sqlite3"),
    }
}

ROOT_URLCONF = "tests.config.test_urls"


TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [
            "templates",
        ],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"



class WebmentionsException(Exception):
    """Base class for mentions-specific exceptions."""

    pass


class TargetWrongDomain(WebmentionsException):
    """Target URL does not point to any domain in settings.ALLOWED_HOSTS."""

    pass


class TargetDoesNotExist(WebmentionsException):
    """Target URL does not point to an object on our server."""

    pass


class SourceNotAccessible(WebmentionsException):
    """Source URL does not exist, or returns an error code."""

    pass


class SourceDoesNotLink(WebmentionsException):
    """Source URL exists but does not contain link to our content."""

    pass


class BadConfig(WebmentionsException):
    """URL resolution completed but did not include required data.

    The returned ResolverMatch.kwargs object must have entries for:
        - 'model_name'
        - 'slug'
    """

    pass


class ImplementationRequired(WebmentionsException):
    """A MentionableMixin model has not implemented a required method."""

    pass


class NotEnoughData(WebmentionsException):
    """Attempted to build a model instance but did not have enough data to do so."""

    pass



"""
Tests for handling webmentions are sent to us from elsewhere.
"""

import logging

from django.http import HttpResponse
from django.test import Client

from mentions.tasks import incoming_webmentions
from mentions.util import split_url
from tests import WebmentionTestCase
from tests.models import MentionableTestModel
from tests.util import testfunc

log = logging.getLogger(__name__)


class _MockHttpClient:
    """Wrap Django Client with API for `python-requests`."""

    def __init__(self, client: Client):
        self.django_client = client

    def get(self, url):
        return _MockResponse(self.django_client.get(url))


class _MockResponse:
    """Wrap Django HttpResponse with API for `python-requests`."""

    def __init__(self, response: HttpResponse):
        self.response = response
        self.headers = response.headers
        self.status_code = response.status_code
        self.text = response.content


class IncomingWebmentionsTests(WebmentionTestCase):
    def setUp(self):
        target_pk, self.target_slug = testfunc.get_id_and_slug()
        self.target_url = testfunc.get_url(self.target_slug)

        MentionableTestModel.objects.create(
            pk=target_pk,
            slug=self.target_slug,
            allow_incoming_webmentions=True,
        )

    def test_get_target_path(self):
        """Ensure that path is retrieved from url correctly."""
        scheme, domain, path = split_url(self.target_url)
        self.assertEqual(testfunc.get_urlpath(self.target_slug), path)

    def test_get_target_object(self):
        """Ensure that database object is retrieved from url correctly."""
        retrieved_model = incoming_webmentions._get_target_object(self.target_url)

        self.assertEqual(retrieved_model.slug, self.target_slug)

    def test_get_incoming_source(self):
        """Ensure that webmention source page can be retrieved correctly."""

        _, source_slug = testfunc.get_id_and_slug()
        source_url = MentionableTestModel.objects.create(
            slug=source_slug
        ).get_absolute_url()

        self.assertIsNotNone(
            incoming_webmentions._get_incoming_source(
                source_url, client=_MockHttpClient(self.client)
            )
        )



"""
View names for use in urlpatterns and reverse lookup.
"""

"""API endpoint which receives new webmentions."""
webmention_api_incoming = "incoming_webmention_api_view"

"""API endpoint for retrieving existing webmentions for some target URL or object."""
webmention_api_get_for_object = "get_webmentions_for_object_api_view"



from mentions.exceptions import ImplementationRequired
from tests import WebmentionTestCase
from tests.models import (
    BadTestModelMissingAllText,
    BadTestModelMissingGetAbsoluteUrl,
    MentionableTestModel,
)
from tests.util import testfunc


def _create_model_instance(Model):
    pk, slug = testfunc.get_id_and_slug()

    return Model.objects.create(
        pk=pk,
        slug=slug,
        content=testfunc.random_domain(),
    )


class MentionableMixinImplementationTests(WebmentionTestCase):
    """MentionableMixin implementation tests"""

    def test_correct_implementation(self):
        """Basic MentionableMixin implementation works correctly."""
        obj = _create_model_instance(MentionableTestModel)

        response = self.client.get(obj.get_absolute_url())
        self.assertEqual(200, response.status_code)
        self.assertEqual(obj.all_text(), obj.content)

    def test_missing_get_absolute_url(self):
        """Mentionable model that does not override get_absolute_url raises ImplementationRequired exception"""
        obj = _create_model_instance(BadTestModelMissingGetAbsoluteUrl)

        with self.assertRaises(ImplementationRequired):
            obj.get_absolute_url()

    def test_unimplemented_alltext(self):
        """Mentionable model that does not override all_text raises ImplementationRequired exception"""
        obj = _create_model_instance(BadTestModelMissingAllText)

        with self.assertRaises(ImplementationRequired):
            obj.all_text()



from django.conf import settings
from django.urls import reverse

from mentions.views import view_names


class WebmentionHeadMiddleware:
    """Automatically add webmention endpoint to HTTP headers of any HttpResponse
    objects created by this server."""

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        port = request.META.get("SERVER_PORT", 80)
        path = reverse(view_names.webmention_api_incoming)
        response["Link"] = (
            f"<{request.scheme}://{settings.DOMAIN_NAME}:{port}{path}>"
            ';rel="webmention"'
        )
        return response



"""
Chunks of markup.
"""

import logging

from tests.util import testfunc

log = logging.getLogger(__name__)

BASE_HTML = """
<!DOCTYPE html>
<html lang="???">
<head>{head}</head>
<body>
{body}
</body>
</html>
"""


def build_html(head=None, body=None) -> str:
    return BASE_HTML.format(head=head, body=body)


def http_header_link(url):
    return f'<{url}>; rel="webmention"'


def _html_head_link(url):
    return f'<link rel="webmention" href="{url}"/>'


def _html_body_link(url):
    return f'<a rel="webmention" href="{url}">Endpoint!</a>'


def http_link_endpoint():
    return http_header_link(testfunc.endpoint_submit_webmention())


def html_head_endpoint():
    return build_html(
        head=_html_head_link(testfunc.endpoint_submit_webmention()),
        body=None,
    )


def html_body_endpoint():
    return build_html(
        head=None,
        body=_html_body_link(testfunc.endpoint_submit_webmention()),
    )


def html_all_endpoints(content):
    return build_html(
        head=_html_head_link(testfunc.endpoint_submit_webmention()),
        body=f"""<div>{content}{_html_body_link(testfunc.endpoint_submit_webmention())}This is arbitrary...</div>""",
    )



import logging

from django.conf import settings
from django.db import models

from .base import MentionsBaseModel
from .mixins.quotable import QuotableMixin

log = logging.getLogger(__name__)


def _approve_default():
    if hasattr(settings, "WEBMENTIONS_AUTO_APPROVE"):
        return settings.WEBMENTIONS_AUTO_APPROVE
    return False


class Webmention(QuotableMixin, MentionsBaseModel):
    """An incoming webmention that is received by your server."""

    sent_by = models.URLField(
        blank=True,
        help_text="Source address of the HTTP request that sent this webmention",
    )

    approved = models.BooleanField(
        default=_approve_default, help_text="Allow this webmention to appear publicly"
    )
    validated = models.BooleanField(
        default=False,
        help_text="True if both source and target have been validated, "
        "confirmed to exist, and source really does link to target",
    )

    notes = models.CharField(max_length=1024, blank=True)

    def approve(self):
        self.approved = True

    class Meta:
        ordering = ["-created_at"]
        permissions = [
            ("approve_webmention", "Can approve received Webmentions for publishing.")
        ]

    def __str__(self):
        return (
            f"{self.source_url} -> {self.target_url} "
            f"[validated={self.validated}, approved={self.approved},"
            f"content_type={self.content_type}, id={self.object_id}]"
        )


class OutgoingWebmentionStatus(MentionsBaseModel):
    """Status tracker for webmentions that you (attempt to) send from your server.

    Used primarily for logging of outgoing mentions.
    """

    source_url = models.URLField(
        help_text="The URL on your server where this mention originates",
    )
    target_url = models.URLField(
        help_text="The URL that you mentioned.",
    )
    target_webmention_endpoint = models.URLField(
        null=True,
        blank=True,
        help_text="The endpoint URL to which we sent the webmention",
    )
    status_message = models.CharField(
        max_length=1024,
        help_text="Success, or an explanation of what went wrong.",
    )
    response_code = models.PositiveIntegerField(default=0)

    successful = models.BooleanField(default=False)

    def __str__(self):
        return (
            f"{self.source_url} -> {self.target_url} "
            f"(endpoint={self.target_webmention_endpoint}): "
            f"[{self.successful}] {self.status_message} "
            f"[{self.response_code}]"
        )

    class Meta:
        ordering = ["-created_at"]
        verbose_name_plural = "Outgoing Webmention Statuses"



"""
Tests for webmentions that originate on our server, usually pointing somewhere else.
"""
import logging
from unittest.mock import Mock, patch

import requests
from django.conf import settings

from mentions.models import OutgoingWebmentionStatus
from mentions.tasks import outgoing_webmentions, process_outgoing_webmentions
from tests import MockResponse, WebmentionTestCase
from tests.util import testfunc

log = logging.getLogger(__name__)


TARGET_DOMAIN = testfunc.random_domain()

OUTGOING_WEBMENTION_HTML = f"""<html>
<head><link rel="webmention" href="/webmention/" /></head>
<body>blah blah 
<a href="https://{TARGET_DOMAIN}/">This is a mentionable link</a> 
blah blah</body></html>
"""

OUTGOING_WEBMENTION_HTML_MULTIPLE_LINKS = f"""<html>
<head><link rel="webmention" href="/webmention/" /></head>
<body>blah blah 
<a href="https://{TARGET_DOMAIN}/">This is a mentionable link</a> 
<a href="https://{TARGET_DOMAIN}/some-article/">This is another mentionable link</a> 
blah blah</body></html>
"""

OUTGOING_WEBMENTION_HTML_NO_LINKS = """<html>
<head><link rel="webmention" href="/webmention/" /></head>
<body>blah blah no links here blah blah</body></html>
"""


def _mock_get_ok(url, headers=None, **kwargs):
    return {
        f"https://{TARGET_DOMAIN}/": MockResponse(
            url,
            text=OUTGOING_WEBMENTION_HTML,
            headers=headers,
            status_code=200,
        ),
        f"https://{TARGET_DOMAIN}/some-article/": MockResponse(
            url,
            text=OUTGOING_WEBMENTION_HTML_MULTIPLE_LINKS,
            headers=headers,
            status_code=200,
        ),
    }.get(url)


def _mock_post_ok(url, headers=None, **kwargs):
    return {
        f"https://{TARGET_DOMAIN}/webmention/": MockResponse(
            url,
            headers=headers,
            status_code=200,
        ),
    }.get(url)


def _mock_post_error(url, headers=None, **kwargs):
    return {
        f"https://{TARGET_DOMAIN}/webmention/": MockResponse(
            url,
            text=OUTGOING_WEBMENTION_HTML,
            headers=headers,
            status_code=400,
        ),
    }.get(url)


def _patch_get(ok: bool):
    return patch.object(requests, "get", Mock(side_effect=_mock_get_ok if ok else None))


def _patch_post(ok: bool):
    return patch.object(
        requests, "post", Mock(side_effect=_mock_post_ok if ok else _mock_post_error)
    )


class OutgoingWebmentionsTests(WebmentionTestCase):
    source_url = f"https://{settings.DOMAIN_NAME}/some-url-path/"

    @_patch_post(ok=True)
    def test_send_webmention(self):
        """_send_webmention should return True with status code when webmention is accepted by server."""

        success, status_code = outgoing_webmentions._send_webmention(
            source_url=self.source_url,
            endpoint=f"https://{TARGET_DOMAIN}/webmention/",
            target=f"https://{TARGET_DOMAIN}/",
        )

        self.assertTrue(success)
        self.assertEqual(200, status_code)

    @_patch_post(ok=False)
    def test_send_webmention__with_endpoint_error(self):
        """_send_webmention should return False with status code when webmention is not accepted by server."""

        success, status_code = outgoing_webmentions._send_webmention(
            source_url=self.source_url,
            endpoint=f"https://{TARGET_DOMAIN}/webmention/",
            target=f"https://{TARGET_DOMAIN}/",
        )

        self.assertFalse(success)
        self.assertEqual(400, status_code)

    @_patch_post(ok=True)
    @_patch_get(ok=True)
    def test_process_outgoing_webmentions(self):
        """Test the entire process_outgoing_webmentions task with no errors."""
        self.assertEqual(0, OutgoingWebmentionStatus.objects.count())

        successful_submissions = process_outgoing_webmentions(
            self.source_url, OUTGOING_WEBMENTION_HTML
        )

        self.assertEqual(1, successful_submissions)
        self.assertEqual(1, OutgoingWebmentionStatus.objects.count())

        successful_submissions = process_outgoing_webmentions(
            self.source_url, OUTGOING_WEBMENTION_HTML_MULTIPLE_LINKS
        )

        self.assertEqual(2, successful_submissions)
        self.assertEqual(3, OutgoingWebmentionStatus.objects.count())

    # No network requests should be made if links not found in text
    @patch.object(requests, "get", None)
    @patch.object(requests, "post", None)
    def test_process_outgoing_webmentions__with_no_links_found(self):
        """Test the entire process_outgoing_webmentions task with no links in provided text."""
        self.assertEqual(0, OutgoingWebmentionStatus.objects.count())

        successful_webmention_submissions = process_outgoing_webmentions(
            self.source_url, OUTGOING_WEBMENTION_HTML_NO_LINKS
        )

        self.assertEqual(0, successful_webmention_submissions)
        self.assertEqual(0, OutgoingWebmentionStatus.objects.count())

    @_patch_get(ok=True)
    @_patch_post(ok=False)
    def test_process_outgoing_webmentions__with_endpoint_error(self):
        """Test the entire process_outgoing_webmentions task with endpoint error."""

        successful_webmention_submissions = process_outgoing_webmentions(
            self.source_url, OUTGOING_WEBMENTION_HTML
        )

        self.assertEqual(0, successful_webmention_submissions)
        self.assertEqual(1, OutgoingWebmentionStatus.objects.count())



"""
Views that are only used for running tests.
"""

import logging

from django.http import HttpResponse
from django.views.generic.base import TemplateView, View

from tests.models import MentionableTestModel
from tests.util import snippets, testfunc

log = logging.getLogger(__name__)


class AllEndpointsMentionableTestView(View):
    """
    Expose webmention endpoint in HTTP header, HTML HEAD and HTML BODY.
    """

    def get(self, request, *args, **kwargs):
        obj = MentionableTestModel.objects.get(slug=kwargs.get("slug"))
        html = snippets.html_all_endpoints(obj.content)
        response = HttpResponse(html, status=200)
        response["Link"] = snippets.http_header_link(
            testfunc.endpoint_submit_webmention_absolute()
        )
        return response


class SimpleNoObjectTestView(View):
    """
    A simple view with no associated Mentionable object. Should be able to accept
    incoming webmentions but there is no
    """

    def get(self, request):
        return HttpResponse(snippets.html_all_endpoints("whatever content"), status=200)


class TemplateTagTestView(TemplateView):
    template_name = "templatetags_example.html"



import logging
import re
from typing import Optional, Set, Tuple
from urllib.parse import urlsplit

import requests
from celery import shared_task
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import URLValidator

from mentions.models import OutgoingWebmentionStatus
from mentions.util import html_parser

log = logging.getLogger(__name__)


STATUS_MESSAGE_TARGET_UNREACHABLE = "The target URL could not be retrieved."
STATUS_MESSAGE_TARGET_ERROR_CODE = "The target URL returned an HTTP error code."
STATUS_MESSAGE_TARGET_ENDPOINT_ERROR = (
    "The target endpoint URL returned an HTTP error code"
)
STATUS_MESSAGE_OK = "The target server accepted the webmention."


def _save_status(
    outgoing_status: OutgoingWebmentionStatus,
    message: str,
    success: bool,
    target_endpoint: Optional[str] = None,
    response_code: Optional[int] = None,
):
    outgoing_status.status_message = message
    outgoing_status.successful = success
    if target_endpoint:
        outgoing_status.target_webmention_endpoint = target_endpoint
    if response_code:
        outgoing_status.response_code = response_code
    outgoing_status.save()


@shared_task
def process_outgoing_webmentions(source_urlpath: str, text: str) -> int:
    """Find links and in the given text and submit a webmention to any servers that support them.

    Spec:
    https://www.w3.org/TR/webmention/#sender-discovers-receiver-webmention-endpoint

    For each link found in text:
        - Check that the link is alive
        - Search for webmention endpoint in:
            - HTTP response headers
            - HTML <head> element
            - HTML <body> element
        - If an endpoint is found, send a webmention notification

    source_urlpath should be the value returned by model.get_absolute_url() -
    it will be appended to settings.DOMAIN_NAME

    Returns:
         Number of outgoing webmentions that were submitted successfully.
    """

    log.info(f"Checking for outgoing webmention links...")
    mentions_attempted = 0
    mentions_sent = 0
    links_in_text = _find_links_in_text(text)

    if not links_in_text:
        log.debug("No links found in text.")
        return 0

    for link_url in links_in_text:
        result = _process_link(source_urlpath, link_url)

        if result is None:
            # No webmention endpoint found
            continue

        mentions_attempted += 1
        if result is True:
            mentions_sent += 1

    if mentions_sent == mentions_attempted:
        log.info(f"Successfully sent {mentions_sent} webmentions")

    elif mentions_attempted:
        log.warning(
            f"Webmention submission errors: {mentions_sent}/{mentions_attempted} submissions were successful"
        )

    else:
        log.info(f"No links with webmentionable endpoints were found.")

    return mentions_sent


def _process_link(source_urlpath: str, link_url: str) -> Optional[bool]:
    """Try to send a webmention for link_url.

    Returns:
        True if a webmention was submitted successfully.

        False if a webmention endpoint was resolved but submission failed.

        None if a webmention endpoint could not be resolved (i.e. the website does not appear to support webmentions).
    """
    outgoing_status = OutgoingWebmentionStatus.objects.create(
        source_url=source_urlpath,
        target_url=link_url,
    )

    # Confirm that the target url is alive
    log.debug(f"Checking url='{link_url}' for webmention support...")
    try:
        response = requests.get(link_url)
    except Exception as e:
        log.warning(f"Unable to fetch url={link_url}: {e}")
        _save_status(outgoing_status, STATUS_MESSAGE_TARGET_UNREACHABLE, success=False)
        return None

    if response.status_code >= 300:
        log.warning(
            f"Mentioned link '{link_url}' returned status={response.status_code}"
        )
        _save_status(
            outgoing_status,
            STATUS_MESSAGE_TARGET_ERROR_CODE,
            success=False,
            response_code=response.status_code,
        )
        return None

    endpoint = _get_absolute_endpoint_from_response(response)
    if endpoint:
        log.debug(f"Found webmention endpoint: '{endpoint}'")
        success, status_code = _send_webmention(source_urlpath, endpoint, link_url)
        if success:
            log.info(f"Webmention submission successful for '{link_url}'")
            _save_status(
                outgoing_status,
                STATUS_MESSAGE_OK,
                target_endpoint=endpoint,
                success=success,
                response_code=status_code,
            )
            return True

        else:
            log.warning(
                f"Webmention submission failed for '{link_url}' [endpoint={endpoint}]"
            )
            _save_status(
                outgoing_status,
                STATUS_MESSAGE_TARGET_ENDPOINT_ERROR,
                target_endpoint=endpoint,
                success=success,
                response_code=status_code,
            )
            return False
    else:
        log.info(f"No webmention endpoint found for url '{link_url}'")


def _find_links_in_text(text: str) -> Set[str]:
    soup = html_parser(text)
    return {a["href"] for a in soup.find_all("a", href=True)}


def _get_absolute_endpoint_from_response(response: requests.Response) -> Optional[str]:
    endpoint = _get_endpoint_in_http_headers(
        response
    ) or _get_endpoint_in_html_response(response)
    abs_url = _relative_to_absolute_url(response, endpoint)
    log.debug(f"Absolute url: {endpoint} -> {abs_url}")
    return abs_url


def _get_endpoint_in_html_response(response: requests.Response) -> Optional[str]:
    return _get_endpoint_in_html(response.text)


def _get_endpoint_in_http_headers(response: requests.Response) -> Optional[str]:
    """Search for webmention endpoint in HTTP headers."""
    try:
        header_link = response.headers.get("Link")
        if "webmention" in header_link:
            endpoint = re.match(
                r'<(?P<url>.*)>[; ]*.rel=[\'"]?webmention[\'"]?', header_link
            ).group(1)
            log.debug(f"Webmention endpoint found in HTTP header: '{endpoint}'")
            return endpoint
    except Exception as e:
        log.debug(f"Unable to read HTTP headers: {e}")


def _get_endpoint_in_html(html_text: str) -> Optional[str]:
    """Search for a webmention endpoint in HTML."""
    a_soup = html_parser(html_text)

    # Check HTML <head> for <link> webmention endpoint
    try:
        links = a_soup.head.find_all("link", href=True, rel=True)
        for link in links:
            if "webmention" in link["rel"]:
                endpoint = link["href"]
                log.debug(f"Webmention endpoint found in document head: {endpoint}")
                return endpoint
    except Exception as e:
        log.debug(f"Error reading <head> of external link: {e}")

    # Check HTML <body> for <a> webmention endpoint
    try:
        links = a_soup.body.find_all("a", href=True, rel=True)
        for link in links:
            if "webmention" in link["rel"]:
                endpoint = link["href"]
                log.debug(f"Webmention endpoint found in document body: {endpoint}")
                return endpoint
    except Exception as e:
        log.debug(f"Error reading <body> of link: {e}")


def _send_webmention(source_url: str, endpoint: str, target: str) -> Tuple[bool, int]:
    payload = {
        "target": target,
        "source": f"https://{settings.DOMAIN_NAME}{source_url}",
    }
    response = requests.post(endpoint, data=payload)
    status_code = response.status_code
    if status_code >= 300:
        log.warning(
            f'Sending webmention to "{endpoint}" '
            f"FAILED with status_code={status_code}"
        )
        return False, status_code
    else:
        log.info(
            f'Sending webmention to "{endpoint}" '
            f"successful with status_code={status_code}"
        )
        return True, status_code


def _relative_to_absolute_url(response: requests.Response, url: str) -> Optional[str]:
    """
    If given url is relative, try to construct an absolute url using response domain.
    """
    if not url:
        return None

    try:
        URLValidator()(url)
        return url  # url is already well-formed.
    except ValidationError:
        scheme, domain, _, _, _ = urlsplit(response.url)
        if not scheme or not domain:
            return None
        return f"{scheme}://{domain}/" f'{url if not url.startswith("/") else url[1:]}'



from django.urls import include, path, re_path

from tests.util import constants, viewname
from tests.views import AllEndpointsMentionableTestView, SimpleNoObjectTestView

urlpatterns = [
    # A page associated with a MentionableMixin model with correct configuration - webmentions linked by model instance.
    path(
        fr"with_correct_config/<slug:slug>",
        AllEndpointsMentionableTestView.as_view(),
        kwargs={
            "model_name": constants.model_name,
        },
        name=viewname.with_all_endpoints,
    ),
    # An arbitrary page with no model association - webmentions are linked by URL.
    path(
        "some-page/",
        SimpleNoObjectTestView.as_view(),
        name=viewname.with_no_mentionable_object,
    ),
    path(f"{constants.namespace}/", include("mentions.urls")),
]



"""Constants that are used in several tests."""
import uuid

"""The root path for mentions.urls.
Generated randomly at test runtime as user may include(mentions.urls) to point to any location."""
namespace = uuid.uuid4().hex

"""Dotted name of the model used for testing."""
model_name = "tests.MentionableTestModel"



"""Models that are only used for testing."""

import logging

from django.db import models
from django.urls import reverse

from mentions.models.mixins.mentionable import MentionableMixin
from tests.util import viewname

log = logging.getLogger(__name__)


class MentionableTestModel(MentionableMixin, models.Model):
    """Basic mentionable model with all required methods implemented."""

    content = models.TextField(blank=True, null=True)

    def get_absolute_url(self):
        return reverse(viewname.with_all_endpoints, args=[self.slug])

    def all_text(self):
        return self.content

    class Meta:
        app_label = "tests"


class BadTestModelMissingGetAbsoluteUrl(MentionableMixin, models.Model):
    """A MentionableMixin model that forgot to implement get_absolute_url()"""

    content = models.TextField(blank=True, null=True)

    def all_text(self):
        return self.content

    class Meta:
        app_label = "tests"


class BadTestModelMissingAllText(MentionableMixin, models.Model):
    """A MentionableMixin model that forgot to implement all_text()"""

    content = models.TextField(blank=True, null=True)

    def get_absolute_url(self):
        return reverse(viewname.with_all_endpoints, args=[self.slug])

    class Meta:
        app_label = "tests"
